from ScopusWp.controller import TopController

import optparse

# TODO: Make a fail proof for the citation comment wordpress not well formed
# TODO: Make the new citations functionality with update date
# TODO: Make update date into the main reference
# TODO: Make error safe for http request pool exceptions wrapper method for requests
